
#include <..\AutoitObject.au3>
#include <String.au3>
#include <Crypt.au3>

Global $__TCPClass_oMyErr = ObjEvent("AutoIt.Error", "_TCPClass_ErrFunc")

; ##############################################################
#cs
	File: TCPClient
#ce

#cs
Class: _Class_TCPClient
	Object that simplifiy the creation of Encrypted TCP Client

Property: server_ip
	IP Adress of the server

Property: server_port
	Port number of the server

Property: max_recv_len
	max # of characters to receive (TCPRecv)

Property: Timeout
	Receiving timeout (in ms).
	
	When this timeout is reached, the buffer is fluched, and the data that was
	being transfered is lost.

Property: Callback_Recv
	Function that is called when data (Binary/String) is received from the server
	
	The function must be as this prototype:
	
	>_Func($s_bData)
	
	$s_bData - Received data (Binary/String)
	
Property: Callback_Receiving
	Function that is called when being receiving data from the server
	
	The function must be as this prototype:
	
	>_Func($BufferLenght)

	$BufferLenght - Lenght of the client's receiving buffer (Not equal to the final lenght/size
		of the received data, it's just an indication)

Property: Callback_LostConnection
	Function that is called when the connection with the server is lost.
	
	Has no parameters
	
Property: Callback_Timedout
	Function that is called when a client is timed out
	
	The function must be as this prototype:
	
	>_Func($BufferLenght_LostData)
	
	$BufferLenght_LostData - Lenght of the buffer when timed out, corresponds to the amount of
		lost data (because buffer is flushed when time out occures)
	
	See <Timeout>
#ce

#cs
Constructor: _CLass_TCPClient
	Class constructor

Parameters:
	$ip - Value to assign to <server_ip>
	$port - Value to assign to <server_port>
	$encrypt_pwd - Data encryption password. Must be the same as in the server.
#ce
Func _Class_TCPClient($ip = "", $port = 0, $encrypt_pwd = "topsecret")
	Local $o = _AutoItObject_Class()
	_Crypt_Startup()
	; --- Creates a random name for the var that will containe the User Key for Crypt function
	Local $key_name = "__TCPClient_hKey_" & @HOUR & "_" & @MIN & "_" & @SEC & "_" & Random(1000, 9999, 1)
	Assign($key_name, _Crypt_DeriveKey($encrypt_pwd, $CALG_RC4), 2)
	; ---
	With $o
		; ### Properties
		.AddProperty("server_ip", $ELSCOPE_PUBLIC, $ip)
		.AddProperty("server_port", $ELSCOPE_PUBLIC, $port)
		.AddProperty("max_recv_len", $ELSCOPE_PUBLIC, 4096)
		.AddProperty("Timeout", $ELSCOPE_PUBLIC, 10000)
		.AddProperty("Callback_Recv", $ELSCOPE_PUBLIC, "")
		.AddProperty("Callback_Receiving", $ELSCOPE_PUBLIC, "")
		.AddProperty("Callback_LostConnection", $ELSCOPE_PUBLIC, "")
		.AddProperty("Callback_Timedout", $ELSCOPE_PUBLIC, "")
		; --- Privates
		.AddProperty("timer", $ELSCOPE_PRIVATE, -1)
		.AddProperty("buffer", $ELSCOPE_PRIVATE, "")
		.AddProperty("socket", $ELSCOPE_PRIVATE, -1)
		.AddProperty("key", $ELSCOPE_PRIVATE, $key_name)
		; ### Methods
		.AddMethod("Connect", "__Class_TCPClient_Connect")
		.AddMethod("Disconnect", "__Class_TCPClient_Disconnect")
		.AddMethod("Send", "__Class_TCPClient_Send")
		.AddMethod("Process", "__Class_TCPClient_Process")
		; --- Privates
		.AddMethod("Cry", "__Class_TCPClient_Crypt", True)
		.AddMethod("DCry", "__Class_TCPClient_UnCrypt", True)
		.AddMethod("CheckBuffer", "__Class_TCPClient_CheckBuffer", True)
		; --- Destructor
		.AddDestructor("__Class_TCPClient_Destructor")
	EndWith
	Return $o.object
EndFunc   ;==>_Class_TCPClient

#cs
Method: Connect
	Try to connect to the server specified by <server_ip> and <server_port>

Parameters:
	$cb_Recv - Value to assign to <Callback_Recv>
	$cb_Receiving - Value to assign to <Callback_Receiving>
	$cb_LostConnection - Value to assign to <Callback_LostConnection>
	$cb_TimedOut - Value to assign to <Callback_Timedout>

Return value:
	Success - Returns main socket identifier. 
	Failure - Returns -1 or 0 and set @error to:
 
	- 1 IPAddr is incorrect.
	- 2 port is incorrect.
	
	windows API WSAGetError return value (see MSDN). 

#ce
Func __Class_TCPClient_Connect($oSelf, $cb_Recv = "", $cb_Receiving = "", $cb_LostConnection = "", $cb_TimedOut = "")
	#forceref $oSelf
	; --- 
	; If connected, then disconnect first
	If $oSelf.socket <> -1 Then $oSelf.Disconnect
	; ---
	Local $ret = TCPConnect($oSelf.server_ip, $oSelf.server_port)
	If $ret <> -1 And Not @error Then
	; Connection succesfull
		$oSelf.Callback_Recv = $cb_Recv
		$oSelf.Callback_Receiving = $cb_Receiving
		$oSelf.Callback_LostConnection = $cb_LostConnection
		$oSelf.Callback_TimedOut = $cb_TimedOut
		; ---
		$oSelf.socket = $ret
		Return SetError(0, 0, $ret)
	Else
	; Connection failed
		$oSelf.socket = -1
		Return SetError(@error, @extended, $ret)
	EndIf
EndFunc   ;==>__Class_TCPClient_Connect

#cs
Method: Disconnect
	Disconnect from the server

Return value:
	Success - Returns 1. 
	Failure - Returns 0 and set @error = windows API WSAGetError return value (see MSDN).
#ce
Func __Class_TCPClient_Disconnect($oSelf)
	#forceref $oSelf
	; ---
	; Close socket and stores error value
	Local $ret = TCPCloseSocket($oSelf.socket)
	Local $err = @error
	; Init .socket member var to -1
	$oSelf.socket = -1
	Return SetError($err, 0, $ret)
EndFunc   ;==>__Class_TCPClient_Disconnect

#cs
Method: Send
	Send data (Binary/String) to the server

Parameters:
	$s_bData - The data to be sent (Binary/String)

Return value:
	Success - Returns number of bytes sent to the server. 
	Failure - Returns 0 and set @error = windows API WSAGetError return value (see MSDN).

Remark:
	If you want to send the ' char, you must send ''
#ce
Func __Class_TCPClient_Send($oSelf, $s_bData)
	#forceref $oSelf
	; ---
	; If not connected, then returns error
	If $oSelf.socket = -1 Then Return SetError(1, 0, 0)
	; ---
	If IsString($s_bData) Then
		$s_bData = StringReplace($s_bData, "'", "''")
	EndIf
	; ---
	Return SetError(@error, @extended, TCPSend($oSelf.socket, $oSelf.Cry($s_bData)))
EndFunc   ;==>__Class_TCPClient_Send

#cs
Method: Process
	Function that calls all the client's callbacks when events occures (Receiving, Lost connection...)

Remark:
	This method must be called in loop.
#ce
Func __Class_TCPClient_Process($oSelf)
	#forceref $oSelf
	; ---
	; If not connected, then returns error
	If $oSelf.socket = -1 Then Return SetError(1, 0, 0)
	; ---
	Local $recv = TCPRecv($oSelf.socket, $oSelf.max_recv_len)
	;ConsoleWrite($recv & " - @error = " & @error & " - @extended = " & @extended & @CRLF)
	If @error Then ; Disconnected
		$oSelf.Disconnect()
		If $oSelf.Callback_LostConnection Then
			Call($oSelf.Callback_LostConnection)
		EndIf
	EndIf
	; ---
	If $recv Then
		$oSelf.timer = TimerInit()
		$oSelf.buffer = $oSelf.buffer & String($recv)		
	EndIf
	; Check timeout
	If $oSelf.buffer Then
		If TimerDiff($oSelf.timer) >= $oSelf.timeout Then
			If $oSelf.Callbakc_TimedOut Then
				Call($oSelf.Callbakc_TimedOut, BinaryLen($oSelf.buffer))
			EndIf
			$oSelf.buffer = ""
		EndIf
	EndIf
	; ---
	If $oSelf.CheckBuffer Then
		Local $ret = _StringBetween($oSelf.buffer, Chr(2), Chr(3))
		For $elem In $ret
			If $oSelf.Callback_Recv Then
				Call($oSelf.Callback_Recv, $oSelf.DCry($elem))
			EndIf
		Next
		$oSelf.buffer = ""
	Else
		If $oSelf.Callback_Receiving And $oSelf.buffer Then
			Call($oSelf.Callback_Receiving, BinaryLen($oSelf.buffer))
		EndIf
	EndIf
EndFunc   ;==>__Class_TCPClient_Recv

; ##############################################################
; ### Internals ################################################
; ##############################################################

Func __Class_TCPClient_Crypt($oSelf, $data)
	#forceref $oSelf
	; ---
	Local $ret
	; ---
	If Not IsBinary($data) Then $data = String($data)
	If IsBinary($data) Then
		$ret = Chr(2) & String(_Crypt_EncryptData($data, Eval($oSelf.key), $CALG_USERKEY)) & Chr(3)
	Else
		$ret = Chr(2) & String(StringTrimLeft(_Crypt_EncryptData($data, Eval($oSelf.key), $CALG_USERKEY), 2)) & Chr(3)
	EndIf
	; ---
	;ConsoleWrite("Crypt: " & $data & " (" & VarGetType($data) & ") => " & $ret & " (" & VarGetType($ret) & ")" & @CRLF)
	Return $ret
EndFunc   ;==>__Class_TCPClient_Crypt

Func __Class_TCPClient_UnCrypt($oSelf, $data)
	#forceref $oSelf
	; ---
	Local $ret
	; ---
	If StringLeft($data, 2) = "0x" Then
		$ret = _Crypt_DecryptData($data, Eval($oSelf.key), $CALG_USERKEY)
	Else
		$ret = BinaryToString(_Crypt_DecryptData(Binary("0x" & $data), Eval($oSelf.key), $CALG_USERKEY))
	EndIf
	; ---
	;ConsoleWrite("DCry: " & $data & " (" & VarGetType($data) & ") => " & $ret & " (" & VarGetType($ret) & ")" & @CRLF)
	Return $ret
EndFunc   ;==>__Class_TCPClient_UnCrypt

Func __Class_TCPClient_CheckBuffer($oSelf)
	#forceref $oSelf
	; ---
	If StringLeft($oSelf.buffer, 1) = Chr(2) And StringRight($oSelf.buffer, 1) = Chr(3) Then
		Return 1
	Else
		Return 0
	EndIf
EndFunc   ;==>__Class_TCPClient_CheckBuffer

; ##############################################################

Func __Class_TCPClient_Destructor($oSelf)
	#forceref $oSelf
	; ---
	_Crypt_DestroyKey(Eval($oSelf.key))
	_Crypt_Shutdown()
	$oSelf.Disconnect()
EndFunc   ;==>__Class_TCPClient_Destructor

; ##############################################################

Func _TCPClass_ErrFunc()
    ConsoleWrite("COM Error, ScriptLine(" & $__TCPClass_oMyErr.scriptline & ") : Number 0x" & Hex($__TCPClass_oMyErr.number, 8) & " - " & $__TCPClass_oMyErr.windescription & @CRLF)
EndFunc    ;==>_ErrFunc
